from PyPFD import PyPFDMarkovTransition

def matrix_mul(A,B):
    resultado = [[0 for _ in range(len(B[0]))] for _ in range(len(A))]

# Multiplicação
    for i in range(len(A)):
        for j in range(len(B[0])):
            for k in range(len(B)):
                resultado[i][j] += A[i][k] * B[k][j]


    return resultado

def markov_cal_Ntest(transitionM,safeV,testsM,testV):

    rM = [[0 for _ in range(len(transitionM))]]
    rM[0][0] = 1
    pfd = 0 
    ntest = len(testV)

    maxT = 0
    for tst in testV:

        if tst > maxT: 
            maxT = tst

    safeM = []

    if len(safeV) > 1:
        safeM.append(safeV)
    else:
        safeM = safeV

    for i in range(maxT):
       rM = matrix_mul(rM,transitionM)

       for k in range(ntest):
          if i % testV[k] == 0:
           rM = matrix_mul(rM,testsM[k])
   
       for j in range(len(rM[0])):
           if safeM[0][j] == 0:
             pfd+=rM[0][j]

    pfd = pfd/maxT

    return {"stateVector": rM, "pfdavg":pfd}

def markov_cal_Ntest_plot(transitionM,safeV,testsM,testV,n_pt):

    rM = [[0 for _ in range(len(transitionM))]]
    rM[0][0] = 1
    pfd = 0 
    ntest = len(testV)

    maxT = 0
    for tst in testV:
        if tst > maxT: 
            maxT = tst


    interval = int(maxT/(n_pt-1))   

    tempo = []
    returnM = rM
    tempo.append(0)

    safeM = []

    if len(safeV) > 1:
        safeM.append(safeV)
    else:
        safeM = safeV

    for i in range(maxT):
       rM = matrix_mul(rM,transitionM)

       for k in range(ntest):
          if i % testV[k] == 0:
           rM = matrix_mul(rM,testsM[k])
   
       for j in range(len(rM[0])):
           if safeM[0][j] == 0:
             pfd+=rM[0][j]

       if i % interval == 0:
            returnM = returnM + rM
            tempo.append(i)  
 

    pfd = pfd/maxT

    return {"stateMatrix": returnM, "pfdavg":pfd, "tempo": tempo}






#Reliabitity Data:
#Generic dictionary for sensors:

def rd_fr_sensor_analog_transmitter_generic():
    return {"name": "Generic Analog Transmitter",       "λ_du":     2e-7, "λ_dd":   1.8e-6, "λ_su":     2e-7, "λ_sd":   1.8e-6, "Type": "B", "SC": 3, "β": 0.040, "βd": 0.020, "HTF": 0, "MTTR": 160}

def rd_fr_sensor_mechanical_switch():
    return {"name": "Generic Mechanical Switch",        "λ_du":     4e-6, "λ_dd":        0, "λ_su":     4e-6, "λ_sd":    4e-13, "Type": "A", "SC": 2, "β": 0.050, "βd": 0.025, "HTF": 0, "MTTR": 100}

#Generic dictionary for Logic Solver:
def rd_fr_ls_sil3_certified_plc_generic():
    return {"name": "Generic SIL3 Certified PLC",       "λ_du":  6.77e-9, "λ_dd":        0, "λ_su":  1.85e-7, "λ_sd":        0, "Type": "B", "SC": 3, "β": 0.050, "βd": 0.025, "HTF": 1, "MTTR": 100}

#Generifc dictionary for final element:

def rd_fr_fe_valve_onoff_generic():
    return {"name": "Generic ON-OFF Failsafe Valve",    "λ_du":  3.37e-6, "λ_dd":        0, "λ_su":  1.38e-6, "λ_sd":        0, "Type": "A", "SC": 3, "β": 0.050, "βd": 0.025, "HTF": 0, "MTTR": 100}

def rd_fr_st_IEC61508_Tableb9():
    return {"name": "IEC61508-6 Table B9",              "λ_du":   0.5e-5, "λ_dd":        0, "λ_su":        0, "λ_sd":        0, "Type": "A", "SC": 3, "β": 0.100, "βd": 0.050, "HTF": 0, "MTTR":   8}

#Generic data for testing:
def rd_pt_sensor_analog_AGAN():
    return {"name":"As Good As New Test", "PTC": 1}

def rd_pt_sensor_analog_transmitter_DeciceMunicator():
    return {"name":"Device Communication Test", "PTC": 0.54 }

def rd_pt_valve_inline_test():
    return {"name":"Inline test and tightness", "PTC": 0.8 }

def rd_pt_partial_stroke_test():
    return {"name":"Partial stroke test", "PTC": 0.7 }





def factorial(n:int) -> float:
    resultado = 1
    for i in range(2, n + 1):
        resultado *= i
    return resultado

def pfd_N2595_avg_MooN(M:int,N:int,λ_du:float,λ_dd:float,β:float,βd:float,T1_month:float,MTTR:float) -> float:
    """
        pfd_N2595_avg_MooN(M:int,N:int,λ_du:float,λ_dd:float,β:float,T1_month:float,MTTR:float) -> float:
        M  = number of detections/actuations necessary for executing the safety function in MooN voting.
        N  = Total number of devices in the safety function in MooN voting.
        λ_du = dangerous undetected failure rate per hours
        λ_dd = dangerous detected failure rate per hours
        β = commum failure in % 
        T1_month = test interval period in months
        MTTR = mean time to repair

        Returns Probability to failure in demand avarege
    """

    TI = T1_month*730
    DUC = λ_du * β
    DUN = λ_du * (1-β)
    DDC = λ_dd * βd    #Modificado
    DDN = λ_dd * (1-βd)  #Modificado
    pfdavg  = (DUC*TI)/2 + DDC*MTTR + (factorial(N)/(factorial(N-M+2)*factorial(M-1)))*((DUN*TI)**(N-M+1)) + (factorial(N)/(factorial(N-M+1)*factorial(M-1)))*((DUN*TI)**(N-M))*DDN*MTTR
    #pfdavg = (DUC*TI)/2 + DDC*MTTR + (factorial(N)/(factorial(N-M+2)*factorial(M-1)))*((DUN*TI)**(N-M+1)) + (factorial(N)/(factorial(N-M+1)*factorial(M-1)))*((DUN*TI)**(N-M))*DDN*MTTR
    return pfdavg


def pfd_N2595_avg_1oo1_1pt(λ_du:float,λ_dd:float,MT_month:float,TI_month:float,PTC:float,MTTR:float) -> float:
    """
        pfd_N2595_avg_1oo1_1pt(λ_du:float,λ_dd:float,PTC:float,TI_month:float,MT_month:float,MTTR:float) -> float

        λ_du = dangerous undetected failure rate per hours
        λ_dd = dangerous detected failure rate per hours
        PTC = partial test coverage for TI interval (0 <= PTC <= 1)
        TI_month = test interval period in months
        MT_month = mission time period in months
        MTTR = mean time to repair

        Returns Probability to failure in demand avarege
    """
    TI = TI_month*730
    MT = MT_month*730
    λTDU = λ_du * PTC
    λNTDU = λ_du*(1-PTC)
    pfdavg = λTDU*TI/2 + λNTDU*MT/2 + λ_dd*MTTR
    return pfdavg

def pfd_N2595_avg_1oo2_1pt(λ_du:float,λ_dd:float,β:float,βd:float,MT_month:float,TI_month:float,PTC:float,MTTR:float) -> float:
    TI = TI_month*730
    MT = MT_month*730
    λTDUC = λ_du * β * PTC
    λNTDUC = λ_du* β * (1-PTC)
    λTDUN = λ_du * (1-β) * PTC
    λNTDUN = λ_du* (1-β) * (1-PTC)
    λDDC = λ_dd* βd #Modificado
    λDDN = λ_dd* (1-βd) #Modificado
    pfdavg = λTDUC*TI/2 + λNTDUC*MT/2 + λDDC * MTTR + 1/3*( (λTDUN* TI) + (λNTDUN*MT))**2 + ((λTDUN*TI)+(λNTDUN*MT))*λDDN * MTTR + (λDDN * MTTR)**2
    return pfdavg

def pfd_N2595_avg_2oo2_1pt(λ_du:float,λ_dd:float,MT_month:float,TI_month:float,PTC:float,MTTR:float) -> float:

    TI = TI_month*730
    MT = MT_month*730
    λTDU = λ_du * PTC
    λNTDU = λ_du*(1-PTC)
    pfdavg = λTDU*TI + λNTDU*MT + 2*λ_dd*MTTR
    return pfdavg


def pfd_N2595_avg_2oo3_1pt(λ_du:float,λ_dd:float,β:float,βd:float,MT_month:float,TI_month:float,PTC:float,MTTR:float) -> float:
    TI = TI_month*730
    MT = MT_month*730
    λTDUC = λ_du * β * PTC
    λNTDUC = λ_du* β * (1-PTC)
    λTDUN = λ_du * (1-β) * PTC
    λNTDUN = λ_du* (1-β) * (1-PTC)
    λDDC = λ_dd* βd #Modificado
    λDDN = λ_dd* (1-βd) #Modificado
    pfdavg = λTDUC*TI/2 + λNTDUC*MT/2 + λDDC * MTTR + ( (λTDUN* TI) + (λNTDUN*MT))**2 + 3*((λTDUN*TI)+(λNTDUN*MT))*λDDN * MTTR + 3*(λDDN * MTTR)**2
    return pfdavg


#def identityMatrix(n: int):
#    """Cria uma matriz identidade de dimensão n×n."""
#    matriz = [[0 for _ in range(n)] for _ in range(n)]
#    for i in range(n):
#        matriz[i][i] = 1
#    return matriz







def markovMatrixDict_1oo1_2pt(λ_du:float, PDC1:float,PDC2:float,MTTR:float=0,λ_dd:float=0,λ_s:float=0):

#0->OK
#1->FS
#2->FDD
#3->FDU1 - Dicovered by Partial Test 1
#4->FDU2 - Dicovered by Partial Test 2
#5->FDU3 - Dicovered by AGAN Test

    λ_du1 = λ_du*PDC1
    λ_du2 = λ_du*PDC2
    λ_du3 = λ_du*(1-PDC1-PDC2)
    mRT = 1/MTTR

    tmatrix = [[-1,λ_s,λ_dd,λ_du1,λ_du2,λ_du3],
              [mRT,-1,0,0,0,0],
              [mRT,0,-1,0,0,0],
              [0,0,0,-1,0,0],
              [0,0,0,0,-1,0],
              [0,0,0,0,0,-1]]
    


    for i in range(0, 5+1):
        tmatrix[i][i] = 1
        for j in range(0,5+1):
            if i != j:
                tmatrix[i][i] -= tmatrix[i][j] 
                
    #print("'''''''''''''''''''''''''")
    #print(tmatrix)
    #print("'''''''''''''''''''''''''")




    testM=  [[1,0,0,0,0,0],
          [0,1,0,0,0,0],
          [0,0,1,0,0,0],
          [1,0,0,0,0,0],
          [1,0,0,0,0,0],
          [1,0,0,0,0,0]]

    testM_pt1= [[1,0,0,0,0,0],
             [0,1,0,0,0,0],
             [0,0,1,0,0,0],
             [1,0,0,0,0,0],
             [0,0,0,0,1,0],
             [0,0,0,0,0,1]]
    
    testM_pt2= [[1,0,0,0,0,0],
             [0,1,0,0,0,0],
             [0,0,1,0,0,0],
             [1,0,0,0,0,0],
             [1,0,0,0,0,0],
             [0,0,0,0,0,1]]

    stateVector = [[1,0,0,0,0,0]]

    safeVector= [[1,1,0,0,0,0]]

    return {"transitionM": tmatrix, "testM": testM, "testM_pt1": testM_pt1, "testM_pt2": testM_pt2, "stateVector":stateVector,"safeVector":safeVector}


def markovMatrixDict_1oo2(λ_du:float,λ_dd:float,λ_su:float,λ_sd:float, β:float, βd:float,  MTTR:float):

    mRT = 1/MTTR

    λ_dun= λ_du*(1-β)
    λ_duc= λ_du*(β)
    λ_ddn= λ_dd*(1-βd)
    λ_ddc= λ_dd*(βd)
    λ_sun= λ_su*(1-β)
    λ_suc= λ_su*(β)
    λ_sdn= λ_sd*(1-βd)
    λ_sdc= λ_sd*(βd)
    λ_s = λ_sd+λ_su



    
    tmatrix = [[-1,    2*λ_ddn,   2*λ_dun,   λ_sdc+λ_suc+2*λ_sdn+2*λ_sun,   λ_ddc,   λ_duc],
              [mRT,         -1,         0,                           λ_s,    λ_dd,       0],
              [  0,          0,        -1,                           λ_s,    λ_dd,    λ_du],
              [mRT,          0,         0,                            -1,       0,       0],
              [mRT,          0,         0,                             0,      -1,       0],
              [0,            0,         0,                             0,       0,      -1]]

    for i in range(0, 5+1):
        tmatrix[i][i] = 1
        for j in range(0,5+1):
            if i != j:
                tmatrix[i][i] -= tmatrix[i][j] 

    testM=  [[1,0,0,0,0,0],
          [0,1,0,0,0,0],
          [0,0,1,0,0,0],
          [1,0,0,0,0,0],
          [1,0,0,0,0,0],
          [1,0,0,0,0,0]]

    stateVector = [[1,0,0,0,0,0]]

    safeVector= [[1,1,1,1,0,0]]

    return {"transitionM": tmatrix, "testM": testM, "stateVector":stateVector,"safeVector":safeVector}


def markovMatrixDict_1oo2_simplified(λ_du:float,λ_dd:float,λ_su:float,λ_sd:float, β:float, βd:float,  MTTR:float):

    µRT = 0
    λDUN= 0
    λDUC= 0
    λDDN= 0
    λDDC= 0
    λSUN= 0
    λSUC= 0
    λSDN= 0
    λSDC= 0
    TI =1

    test_matrix = [
    [0.0, 2*λDDN, λSDC+λSUC+2*λSDN+2*λSUN, 2*λDUN, λDDC, λDUC],
    [µRT, 0.0, 0.0, 0.0, λDDN, 0.0],
    [µRT, 0.0, 0.0, 0.0, 0.0, 0.0],
    [TI, 0.0, 0.0, 0.0, 0.0, λDUN],
    [µRT, 0.0, 0.0, 0.0, 0.0, 0.0],
    [TI, 0.0, 0.0, 0.0, 0.0, 0.0],
    ]




    µRT = 1/MTTR
    λDUN= λ_du*(1-β)
    λDUC= λ_du*(β)
    λDDN= λ_dd*(1-βd)
    λDDC= λ_dd*(βd)
    λSUN= λ_su*(1-β)
    λSUC= λ_su*(β)
    λSDN= λ_sd*(1-βd)
    λSDC= λ_sd*(βd)
    TI =0


    transition_matrix = [
    [0.0, 2*λDDN, λSDC+λSUC+2*λSDN+2*λSUN, 2*λDUN, λDDC, λDUC],
    [µRT, 0.0, 0.0, 0.0, λDDN, 0.0],
    [µRT, 0.0, 0.0, 0.0, 0.0, 0.0],
    [TI, 0.0, 0.0, 0.0, 0.0, λDUN],
    [µRT, 0.0, 0.0, 0.0, 0.0, 0.0],
    [TI, 0.0, 0.0, 0.0, 0.0, 0.0],
    ]

    vector_of_states = [1, 1, 1, 1, 0, 0]
    vector_of_names = ['OK', 'Degrade 1DD', 'Degraded 2S', 'Degraded 1DU', 'Failure 2DD', 'Failure 2DU']



    for i in range(0, 5+1):
        transition_matrix[i][i] = 1
        for j in range(0,5+1):
            if i != j:
                transition_matrix[i][i] -= transition_matrix[i][j] 


    stateVector = [[1,0,0,0,0,0]]

    return {"transitionM": transition_matrix, "testM": test_matrix, "stateVector":stateVector,"safeVector":vector_of_states} 
